td, th {
    font-family: Verdana, Helvetica, sans-serif;
    margin: 0;
    padding: 0 1ex;
}
td {
    border-width: 1px;
    border-style: solid;
}
table {
    border-spacing: 0;
}
.shortlink {
    font-size: smaller;
}
